# $RCSfile: oof.py,v $
# $Revision: 1.155 $
# $Author: rlua $
# $Date: 2006/03/15 23:14:28 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

import sys, os, types, string, random, code

# g++ code won't correctly resolve dynamic casts in shared libraries
# unless RTLD_GLOBAL is set when the library is loaded.  See
# http://gcc.gnu.org/faq.html#dso.  This must be done before any other
# oof modules are loaded.
try:
    sys.setdlopenflags(0x101)          # RTLD_GLOBAL (0x100) | RTLD_LAZY (0x001)
except AttributeError:
    pass

# The following calls initialize elements of the C++/Python interface,
# and must done before anything else that might call OOF C++ code.
import oof2.SWIG.common.switchboard          # message passing

from oof2.SWIG.common import crandom
from oof2.SWIG.common import lock
from oof2.SWIG.common import ooferror
from oof2.common import debug
from oof2.SWIG.common import devel
from oof2.common import parallel_enable
from oof2.common import oof_getopt as getopt
from oof2.common import subthread
from oof2.common import thread_enable
from oof2.common import utils
from oof2.common import oofversion
from oof2.common.IO import reporter
from oof2.common.IO import automatic

########################################

# Option processing:

# Recognized OOF options are processed and removed from sys.argv by
# the remove_option function.  Special options are provided for
# passing arguments through to some oof dependencies.  These work like
# "--non_oof_options="non_oof_opt1 ...".  We also use a custom version
# of getopt, where unrecognized options are NOT AN ERROR -- they
# remain in sys.argv, and may be consumed by various dependencies (in
# particular, mpi, but also gtk).  This subsequent processing may
# result in error messages which appear to come from OOF, but actually
# originate with dependencies.


# Non-argument options are easy.  Argument options can have the form
# where sys.argv[i] is the option flag, and sys.argv[i+1] is the
# argument, or alternatively may only occur in sys.argv[i], with an
# equal sign, as "item=argument".
def remove_option(item, argument):
    for i in range(len(sys.argv)):
        if sys.argv[i]==item:
            if argument=="":  # Options without arguments.
                del sys.argv[i]
                break
            # If there is an argument, find it in either form.
            elif i<len(sys.argv) and sys.argv[i+1]==argument:
                del sys.argv[i]
                del sys.argv[i] # First del changes index of second entry.
                break
        elif sys.argv[i]==item + "=" + argument:
            del sys.argv[i]
            break
        

def state_options_and_quit():
    main_options_string = """
This is oof2 version %s.
Usage: oof2 [options]
The options are:
option      argument    description
--------------------------------------------------------------
--text                   Turn off graphics mode
--help                   Display valid options and exit
--version                Display version number and exit
--gtk=      gtk options  Extra options for graphics mode
--seed=     integer      Provide a random number seed
--quiet                  Quit quietly when done
--batch                  Quit immediately after running scripts (implies --text)
--debug                  Turn on debugging mode
--unthreaded             Don't use multiple execution threads.""" \
    % oofversion.version
    
    devel_options_string = """
--parallel               Start-up parallel processing mode """

    data_options_string = """
    
The following options may be present more than once:
--script=   file         Load a script
--image=    file         Load an image
--data=     file         Load a data file
--import=   module       Import a python extension module
Extension modules are loaded first, and then the script, image, and
data files are loaded in the order that they are specified.
""" 
    print main_options_string,
    if devel.devel()>=1:
        print devel_options_string,
    print data_options_string
    sys.exit(1)

def state_version_and_quit():
    print "This is oof2 version %s." % oofversion.version
    sys.exit(1)

##

startupfiles = []
startupimports = []
gtk_options = None
text_mode = 0
seeded = 0
help_mode = None
version_mode = None
batch_mode = 0
socket_mode = 0

def process_inline_options():
    # Defaults for option switches.
    global startupfiles
    global startupimports
    global gtk_options
    global text_mode
    global seeded
    global help_mode
    global version_mode
    global batch_mode
    option_list = ['text', 'help', 'version', 'quiet', 'batch',
                   'gtk=', 'unthreaded', 'socket=', 'script=', 'seed=',
                   'data=', 'image=', 'import=', 'debug']
    if devel.devel()>=1:
        option_list += ['parallel']
    try:
        (optlist, args) = getopt.getopt(sys.argv[1:], '', option_list)
    except getopt.error, message:
        print message
        state_options_and_quit() ## exits if malformed arguments found
    for opt in optlist:
        if opt[0] == '--gtk':
            gtk_options = opt[1]
            remove_option(opt[0],opt[1]) ## options of the type -option
        elif opt[0] == '--parallel':
            parallel_enable.set(1)
            remove_option(opt[0],'') ## options of the type -option
        elif opt[0] in ('--unthreaded',):
            thread_enable.set(0)
            lock.disable_all()
            remove_option(opt[0],'') ## options of the type -option            
        elif opt[0] in ('--text',):
            text_mode = 1 # Anything non-zero will work.
            remove_option(opt[0],'') ## options of the type -option
        elif opt[0] in ('--help',):
            help_mode = 1
            remove_option(opt[0],'')
        elif opt[0] in ('--version',):
            version_mode = 1
            remove_option(opt[0], '')
        elif opt[0] in ('--script',):
            startupfiles.append(StartUpScript(opt[1]))
            remove_option(opt[0],opt[1]) ## options of the type -option
        elif opt[0] in ('--image',):
            startupfiles.append(StartUpImage(opt[1]))
            remove_option(opt[0],opt[1]) ## options of the type -option
        elif opt[0] in ('--data',):
            startupfiles.append(StartUpData(opt[1]))
            remove_option(opt[0],opt[1]) ## options of the type -option
        elif opt[0] in ('--import',):
            startupimports.append(opt[1])
            remove_option(opt[0], opt[1])
        elif opt[0] in ('--debug',):
            debug.set_debug_mode()
            remove_option(opt[0],'') ## options of the type -option
        elif opt[0] in ('--quiet',):
            from oof2.common import quit
            quit.set_quiet()
            remove_option(opt[0],'') ## options of the type -option
        elif opt[0] in ('--batch',):
            batch_mode = 1
            text_mode = 1
            remove_option(opt[0],'')
        elif opt[0] == '--seed':
            seeded = 1
            random.seed(int(opt[1]))
            crandom.rndmseed(int(opt[1]))
            ## Identified option is removed from sys.argv    
            remove_option(opt[0],opt[1]) ## options of the type -option
    if help_mode:
        state_options_and_quit()
    if version_mode:
        state_version_and_quit()
    
    if gtk_options :
        if not text_mode:
            sys.argv.append(gtk_options)
            ## gtk commands are eaten upon importing module
        else:
            ## send warning message that --gtk was selected and --text was not
            print "gtk options are ignored in text mode"
    else:
        import oof2.SWIG.common.argv
        oof2.SWIG.common.argv.init_argv(sys.argv[1:])
    
    
        
## ######### Notes to ALL developers ######### 
## front_end() is the old run() in serial mode 
##
## If oof is compiled with an MPI implementation and --parallel is selected
## mpi is initialized at run time.
##
## If MPI was compiled but you run oof as you usually
## do, then only the front end is loaded and the back_end()
## remains *dormant*. MPI will be initialized.
##
## If MPI was not compiled, only the front end
## runs and everything works as usual. MPI will not be initialized.
##
## oof runs in parallel by typing:
##
## mpirun -np <N> <more MPI options> oof2 --graphics --parallel <more oof options>
##
## where <N> is the number of processors.
##
## if <N> > 1:
## the back end is loaded, and the new virtual machine
## waits for instructions to be executed.
## If <N> == 1 oof will run as usual, and the back end
## will NOT be loaded.

        
def front_end(no_interp=None):
    global startupfiles
    global gtk_options
    global text_mode
    global seeded
    ## From here on is the serial version.
    if debug.debug() and not seeded:
        random.seed(17)                 # default random seed
        crandom.rndmseed(17)
        
    # When loading modules, use utils.OOFexec so that names are
    # imported into the oof environment, not the oof.run environment.
    if not text_mode:
	# The gtk import dance described below doesn't work when the program
        # has been packaged by cx_freeze.
        # TODO LATER: is checking frozen required for gtk2?
        frozen = hasattr(sys, 'frozen')
	if not frozen:
            import pygtk
            pygtk.require("2.0")
            import gtk
            msg = gtk.check_version(2, 6, 0)
            if msg:
                print msg
                sys.exit()

        import oof2.common.IO.GUI.initialize
        import oof2.engine.IO.GUI.initialize
        import oof2.image.IO.GUI.initialize
        import oof2.tutorials.initialize
    else:                               # text mode
        import oof2.common.initialize
        import oof2.engine.initialize
        import oof2.image.initialize
    import oof2.EXTENSIONS.initialize

    for module in startupimports:
        exec('import ' + module)

    if not text_mode: ## only front end should ever start the GUI
        reporter.report("Welcome to OOF2 version %s!" % oofversion.version)
        # Importing from the GUI modules is forbidden in all files
        # outside of the GUI modules, except for right here.
        from oof2.common.IO.GUI import oof_mainiteration
        ## The files to be loaded must be loaded *after* the GUI
        ## starts, but this routine doesn't regain control once it
        ## starts the GUI. So we have to install the file loader
        ## (loadStartUpFiles) as an idle callback, which will run on
        ## the main thread.  loadStartUpFiles just issues menu
        ## commands that load the files, and if it runs on the main
        ## thread those menu commands will run by Workers on
        ## subthreads, and won't be run sequentially.  So, instead of
        ## installing loadStartUpFiles as an idle callback, we install
        ## subthread.execute and have it call loadStartUpFiles, since
        ## workers on subthreads don't create additional subthreads to
        ## run their menu items.
        if startupfiles:
            oof_mainiteration.oof_idle_add(subthread.execute,
                                           (loadStartUpFiles, (startupfiles,)))
        if not no_interp: # Default case, run on local thread.
            run_gui()
        else:
            # TODO LATER: The gui and no_interp combination is
            # thinkable, but has problems.  You have to run the GUI on
            # a separate thread, but then exceptions show up as modal
            # dialog boxes in the GUI, and block the menu items which
            # raised them, causing a loss of control.
            print "GUI no_interp mode not implemented.  Sorry."
            
    else:                               # text mode
        from oof2.common import quit
        if startupfiles:
            loadStartUpFiles(startupfiles)
            if batch_mode:
                quit.set_quiet()
                quit.quit()
                sys.exit()
        # Make sure we run in text mode even if "-i" flag was not
        # given to the python interpreter.  Format the banner for the
        # current line width.
        if not quit.quiet():
            width = utils.screenwidth()
            wiggles = "//=*=\\\\=*="
            nwiggles = (width-2)/len(wiggles)
            welcome = "Welcome to OOF2 version %s!" % oofversion.version
            nblanks = (width - len(welcome))/2
            banner = wiggles*nwiggles + "//\n\n" \
                     + " "*nblanks + welcome + "\n" + \
                     string.join(utils.format(banner1, width),"\n") + \
                     "\n\n" +  wiggles*nwiggles + "//\n" + \
                     string.join(utils.format(banner2, width), "\n")
        else:
            banner = ""
            
        if not no_interp:
            import code
            import readline
            # Start up the interpreter in the __main__ namespace.
            # This is the namespace that utils.OOFeval and OOFdefine
            # use.  It's not necessarily *this* namespace.
            interp = code.InteractiveConsole(
                sys.modules['__main__'].__dict__)
            interp.interact(banner)
        else: # In no_interp text-mode case, allow exceptions out.
            from oof2.common import excepthook
            excepthook.restore_excepthook()
            from oof2.common import worker
            worker.propagate_exceptions=1

                

# Encapsulated commands to run the GUI and exit cleanly.  Can be
# run on a thread in no_interp mode.
def run_gui():
    from oof2.common.IO.GUI import oofGUI
    oofGUI.start()
    from oof2.common import quit
    quit.reallyquit()
    


banner1 = """
OOF2 was written at the National Institute of Standards and Technology (NIST).
NIST assumes no responsibility for the operation, modification, or mainentance of OOF2."""

banner2 = """
Type "OOF.Copyright()" to see the copyrights associated with OOF2.
Type "OOF.Credits()" for a list of the authors.
Type "help()" for Python help."""

###########################

class StartUpFile:
    def __init__(self, filename):
        self.filename = filename
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.filename)

class StartUpScript(StartUpFile):
    def load(self):
        mainmenu.OOF.File.Load.Script(filename=self.filename)

class StartUpScriptNoLog(StartUpFile):
    def load(self):
        mainmenu.OOF.haltLog()
        mainmenu.OOF.File.Load.Script(filename=self.filename)
        mainmenu.OOF.resumeLog()

class StartUpImage(StartUpFile):
    def load(self):
        basename = os.path.basename(self.filename)
        mainmenu.OOF.Microstructure.Create_From_ImageFile(
            filename=self.filename,
            microstructure_name=basename,
            height=automatic.automatic, width=automatic.automatic)

class StartUpData(StartUpFile):
    def load(self):
        mainmenu.OOF.File.Load.Data(filename=self.filename)

def loadStartUpFiles(files):            # Files is a list of StartUpFile objects
    for file in files:
        file.load()

## identifiers and other global variables for parallel machine live here.

_rank = 0


def start_parallel_machine():
    global _rank
    try:
        from oof2.SWIG.common import mpitools
    except ImportError:
        pass
    else:
        mpitools.mpi_initialize(sys.argv)
        _rank = mpitools.rank()



## sockets initialization front and back end
def start_sockets_Front_End():
    if parallel_enable.query() == 1:
        try:
            from oof2.SWIG.common import mpitools
        except ImportError:
            raise ooferror.ErrSetupError(
                "Parallel option requested, but parallel code not present.")
        from oof2.common.IO import socket2me
        ## Tell back end what port to listen
        s_name = mpitools.get_processor_name()
        s_address = socket2me.socketPort.getPort()
        mpitools.broadcast_string("%i"%s_address, 0)
        mpitools.broadcast_string(s_name, 0)
        ## Tell back end the number of processors on the back end
        socket2me.socketPort.connect(mpitools.size()-1)        
def start_sockets_Back_End():
    if parallel_enable.query() == 1:
        from oof2.common.IO import socket2me
        from oof2.SWIG.common import mpitools
        ## receive from front end what port to listen
        s_address = mpitools.receive_broadcast_string(0)
        s_name = mpitools.receive_broadcast_string(0)
        ## create a unique SocketInput object
        socket2me.makeSocketInput(s_name, int(s_address), mpitools.rank())




# Main routine.  Takes a "no_interp" argument, which, if not None,
# suppresses running the command interpreter.  This is meant to be
# used by "enclosing" Python routines which import oof and want to
# issue commands to it.  The appropriate way to do this is:
# from oof2.common import oof
# oof.run(no_interp=1)
# For the moment, this only works in text mode.
def run(no_interp=None):
    global _rank
    global startupfiles

    # Look for .oof2rc in the users home directory.
    oofrcpath = os.path.join(os.path.expanduser("~"), ".oof2rc")
    if os.path.exists(oofrcpath):
        startupfiles = [StartUpScriptNoLog(oofrcpath)]+startupfiles

    process_inline_options() ## execute well-formed oof options

    # Import mainmenu only *after* processing command line options, so
    # that the options can affect which menus are loaded.
    global mainmenu
    from oof2.common.IO import mainmenu

    if devel.petsc():
        print "Going to InitPETSc"
        from oof2.SWIG.engine.PETSc.petsc_solverdriver import InitPETSc
        InitPETSc(sys.argv)
        for s in sys.argv:
            print s

    start_parallel_machine() ## start parallel suite (if available)
    if _rank == 0:
        # Open channels to the back end, if appropriate.
        start_sockets_Front_End()
        # Start the GUI, if appropriate, and talk to the user.
        front_end(no_interp) 
    else:
        # Parallel back-end case.

        # MPI does not send the same arguments to back-ends, so
        # explicitly inform the back-end that it's parallel.
        parallel_enable.set(1)
        
        thread_enable.set(0)
        ## debug.set_debug_mode()
        ## initialize back-end
        from oof2.common import quit
        quit.set_quiet() # Back-end exits quietly.

        # Set up the sockets, then wait for commands to arrive.
        start_sockets_Back_End() 
        from oof2.common import backEnd
        backEnd.back_end() 
