// -*- C++ -*-
// $RCSfile: femesh.h,v $
// $Revision: 1.69 $
// $Author: langer $
// $Date: 2006/03/13 20:26:06 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef FEMESH_H
#define FEMESH_H

class ElementIterator;
class FuncNodeIterator;
class FEMesh;
class NodeIterator;
class LinearSolver;

#include <Python.h>

#include "common/lock.h"
#include "common/timestamp.h"
#include "common/pythonexportable.h"
#include "engine/nodalfluxes.h"
#include "engine/outputval.h"
#include <string>
#include <vector>
#include <map>

class CMicrostructure;
class CompoundField;
class Coord;
class CSCPatch;
class CSRmatrix;
class DegreeOfFreedom;
class Element;
class Equation;
class Field;
class Flux;
class FuncNode;
class Material;
class MasterCoord;
class NodalEquation;
class NodalSCPatches;
class Node;
class StiffnessMatrix;
class SubProblem;

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// MeshIterators

enum NodeIteratorType {NI_ALL, NI_SELECTED, NI_UNSELECTED};
enum ElementIteratorType {EI_ALL, EI_SELECTED, EI_UNSELECTED};

class NodeIterator {
private:
  const FEMesh * const mesh;
  NodeIteratorType type;
  std::vector<Node*>::size_type index;
public:
  NodeIterator(const FEMesh * const, NodeIteratorType type=NI_ALL);
  void operator++();
  bool begin() const;
  bool end() const;
  Node *node() const;
  friend class FEMesh;
};

class FuncNodeIterator {
private:
  const FEMesh * const mesh;
  NodeIteratorType type;
  std::vector<FuncNode*>::size_type index;
public:
  FuncNodeIterator(const FEMesh * const, NodeIteratorType type=NI_ALL);
  void operator++();
  bool end() const;
  FuncNode *node() const;
  friend class FEMesh;
};

class ElementIterator {
private:
  const FEMesh * const mesh;
  std::vector<Element*>::size_type index;
  ElementIteratorType type;
public:
  ElementIterator(const FEMesh* const, ElementIteratorType type=EI_ALL);
  void operator++();
  bool end() const;
  int size() const;
  Element *element() const;
  friend class FEMesh;
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

class FEMesh : public PythonExportable<FEMesh> {
  // FEMesh has to be derived from PythonExportable so that it can be
  // used as an argument in switchboard messages.
private:
  CMicrostructure * microstructure;
  RWLock * rwlock;
public:
  FEMesh();
  FEMesh(CMicrostructure *); //for relaxation algorithm only
  virtual ~FEMesh();
  CMicrostructure * get_microstructure() const {return microstructure;}
  void set_microstructure(CMicrostructure * mic) {microstructure = mic;}
  virtual const std::string &classname() const;
  virtual const std::string &modulename() const;

  Node *operator[](const NodeIterator &) const;
  FuncNode *operator[](const FuncNodeIterator &) const;
  Element *operator[](const ElementIterator&) const;

  ElementIterator element_iterator(ElementIteratorType = EI_ALL) const;
  NodeIterator node_iterator(NodeIteratorType = NI_ALL) const;
  FuncNodeIterator funcnode_iterator(NodeIteratorType = NI_ALL) const;


  Node *newMapNode(const Coord&); // the only way to make a Node
  FuncNode *newFuncNode(const Coord&); // the only way to make a FuncNode
  void reserveFuncNodes(int n);	// reserve space for n func nodes
  void reserveMapNodes(int n);	// reserve space for n map nodes
  void AddElement(Element*);
  void reserveElements(int n);

  // Caution: NodeIterator::index is not necessarily the same as
  // node.index().  The argument to FEMesh::getNode() is the
  // NodeIterator::index.
  Node *getNode(int) const;

  // Temporary function for finding the closest node.
  Node *closestNode(const double x, const double y);

  DegreeOfFreedom *CreateDoF(double x=0); // only way to make a DoF
  void RemoveDoF(DegreeOfFreedom*);
  void activate_active_dofs(FuncNode *) const;

  void dump_dof() const;	// for debugging
  void dump_eqn() const;

  // NodalEquations represent one component of one Equation at a Node.
  // They are to the rows of the stiffness matrix what the
  // DegreeOfFreedoms are to the columns.
  NodalEquation *CreateNodalEqn();
  void RemoveNodalEqn(NodalEquation*);

//   void precompute();
  void requirePrecompute();	// call when Materials change or Fields defined
  Lock precomputeLock;
  bool precomputeRequired;

  void make_linear_system();
  const StiffnessMatrix &stiffness_matrix() const;
  const StiffnessMatrix &conductivity_matrix() const;
  const StiffnessMatrix &mass_matrix() const;
  VECTOR_D &boundary_rhs() const;
  void boundary_rhs_init(int);
  void dump_boundary_rhs();
  

  // Run post-solver processing routines in those materials/properties
  // that have them.  This should eventually depend which auxiliary
  // fields are contributed to by the material.
  void post_process();

  std::vector<int> freedofmap;
  std::vector<int> fixeddofmap;
  std::vector<int> indepeqnmap;
  std::vector<int> depeqnmap;
  void rcmapper(bool (FEMesh::*)(int) const, std::vector<int>&, int) const; 
  void mapdofeqs();  // Creates the matrix-partitioning maps.
  void resetdofeqs(); // Undoes BC invocation, sets dofs/eqns to free/indep.
  void cleanlists();
  CSRmatrix Amatrix() const;
  CSRmatrix Bmatrix() const;
  CSRmatrix Cmatrix() const;
  CSRmatrix Cpmatrix() const;
  std::vector<int> A_rowblocks() const;
  std::vector<int> A_colblocks() const;

  Element *getElement(int i) const; //{ return element[i]; }

  void refreshMaterials(PyObject *skeleton);

  int nnodes() const;
  int nelements() const;

  void activate_equation(Equation &);
  void deactivate_equation(Equation &);
  bool is_active_equation(const Equation&) const;
  int active_equation_index(const Equation&) const;
  int n_active_eqns() const; // { return n_active_eqn; }
  //kinetic API starts here
  void kinetic_activate_equation(Equation &);
  void kinetic_deactivate_equation(Equation &);
  bool is_kinetically_active_equation(Equation&);
  bool kinetics_activated();
  //kinetic API ends here

  //dynamics API starts here
  void activate_dynamics(Equation &);
  void deactivate_dynamics(Equation &);
  bool is_dynamically_active_equation(Equation&);
  bool dynamics_activated();
  //dynamics API ends here
  void activate_flux(const Flux&);
  void deactivate_flux(const Flux&);
  bool is_active_flux(const Flux&) const;

  // activate, etc, for Fields have to be routed through Field virtual
  // functions, because they have to know if the Field is a
  // CompoundField or not.  The routines that do the real work are
  // FEMesh::do_activate, etc.
  void activate_field(Field&);
  void deactivate_field(Field&);
  void acquire_field_data(Field &field, const FEMesh*);
  Field &define_field(Field&);
  void undefine_field(Field&);
  void set_in_plane(const Field&, bool);
  bool in_plane(const Field&) const;

  // active fields are being solved for
  void do_activate_field(Field&);
  void do_deactivate_field(Field&);
  bool is_active_field(const Field&) const;
  int n_active_fields() const;
  //  int n_active_dofs() const { return n_active_dof; } // # active dofs per node
  // defined fields have values at nodes
  void do_define_field(Field&);
  void do_undefine_field(Field&);
  bool is_defined_field(const Field&) const;
  int defined_index(const Field&) const; // offset in node's dof list
  int n_defined_dofs() const;// { return n_defined_dof; } // # of dofs per node

  // List of this mesh's defined fields.
  std::vector<CompoundField*>* all_fields() const;
  std::vector<Equation*>* all_equations() const;
  std::vector<Flux*>* all_fluxes() const;
  std::vector<Flux*> allFluxes() const;  // no need to free memory.

  // create & add a new CSCPatch pointer
  void init_scpatches(const std::vector<int>*);
  void add_scpatch(const int, const Material*, 
		   const int,
		   const std::vector<int>*,
		   const std::vector<int>*,
		   const int);
  // getting elements & nodes from the patch
  std::vector<int> *get_elements_from_patch(const int, const Material*);
  std::vector<int> *get_nodes_from_patch(const int, const Material*);
  // recovering flux(es)
  void init_nodalfluxes();
  void recover_fluxes();
  // adding recovered flux
  void add_this_flux(const Material*, const Flux*,
		     const Node*, VECTOR_D*);
  // recovered flux at a given point
  VECTOR_D *get_recovered_flux(const Flux*, const Element*,
			       const MasterCoord&);
  // reporting recovered fluxes at a given point -- debug purpose
  void report_recovered_fluxes(const Element*, const Coord*);
  // estimating error
  double zz_L2_estimate(const Element*, const Flux*);
  void zz_L2_estimate_sub(const Element*, const Flux*,
			  const int&, double&, double&,
			  const MasterCoord&, const double&);
  std::vector<double> *zz_L2_weights(const Flux*,
				     const double&, const double&);
  void zz_L2_weights_sub(const Element*, const Flux*,
			  const int&, double&,
			  const MasterCoord&, const double&);

  // API for setting/refrering to the read-write lock.  Set_rwlock
  // should be called exactly once when the femesh is inserted into a
  // mesh context object.
  void set_rwlock(RWLock *rw) { rwlock = rw; };
  inline RWLock  * get_rwlock() { return rwlock;};

private:
  // These lists can be accessed through the MeshIterators.
  std::vector<FuncNode*> funcnode;// nodes at which dofs are defined
  std::vector<Node*> mapnode;	// nodes only used for geometry mapping
  std::vector<Element*> element;
  int ncount;

  // Storage for CSCPatch's (keyed an assembly node)
  // NodalSCPatches contains as many CSCPatches as no. of Materials
  // at the assembly node.
  std::map<const int, NodalSCPatches*> scpatches;
  // Storage for SCpatch Recovered Fluxes
  std::map<const int, NodalFluxes*> recovered_fluxes;

  // Master lists of degrees of freedom and nodal equations
  std::vector<DegreeOfFreedom*> dof;
  std::vector<NodalEquation*> nodaleqn;
  void housekeeping();		     // do garbage collection on the lists
  bool dof_list_needs_cleaning;
  bool nodaleqn_list_needs_cleaning; // do the lists need to be cleaned up?
  void cleanmap(std::vector<int>&);       // Clean up a particular map.
  // Above routine is called by public method "cleanlists".

  // Is a Flux active?  Stored as an int, because a Flux can be
  // activated more than once.
  std::vector<int> active_flux;
  
  //private machinery to symmetrize the stiffness matrix
  std::vector<int> *global_equation_map;
  /* 
     sends the dof indices from the
     symmetric matrix to the unsymmetric
     (as stored, original) matrix system
  */
  // std::vector<int> * local_rows_mapping; 
  /* 
     proposed row ordering in 
     stiffness matrix
  */

  // This is data that can't live in the Field class, because there is
  // only one instance of each Field, but it might have different
  // properties on each FEMesh, and there might be more than one FEMesh.
  class FieldData {
  public:
    FieldData();
  private:
    //    int active_index;		// used to order the active dofs
    int defined_index;		// used to compute location in a node's doflist
    bool active;		// is this field being solved for?
    bool defined;		// have DoF's been assigned?
    bool in_plane;		// is this field in plane?
    friend class FEMesh;
  };
  std::vector<FieldData> fielddata;

  // This is data that can't live in the Equation class, for the same
  // reason that the FieldData can't live in the Field class.
  class EquationData {
  public:
    EquationData();
  private:
    int active_index;
    bool active;
    bool kin_active;
    bool dyn_active;
    friend class FEMesh;
  };
  std::vector<EquationData> eqndata;
  void clean_nodaleqn();
  void clean_doflist();

  int n_defined_dof;		// number of dofs defined at each node
  //  int n_active_dof;		// number of active dofs at each node
  int n_active_eqn;		// number of active eqns at each node
  int n_active_field;		// number of active fields at each node


  StiffnessMatrix *stiffness;
  VECTOR_D *boundary_right_side;  // Boundary's contributions to forces.
  VECTOR_D *body_force_rhs;       // RHS contribs from integrating properties

  // Routines used to pick rows and columns of the stiffness matrix
  // when extracting submatrices
  bool solve_for_this_dof(int) const;
  bool solve_this_nodal_eqn(int) const;
  bool dont_solve_for_this_dof(int) const;
  bool dont_solve_this_nodal_eqn(int) const;

  // Routines for extracting parts of the rhs and solution vectors
public:
  StiffnessMatrix * get_stiffness() const {return stiffness;}
  void get_unknowns(VECTOR_D&, const std::vector<int>&) const;
  void get_rhs(VECTOR_D&, const std::vector<int>&) const;
  // The opposites of the above routines...
  void set_unknowns(const VECTOR_D&, const std::vector<int>&);
  void set_rhs(const VECTOR_D&, const std::vector<int>&);

  // Functions for symmetrizing the stiffness matrix, if
  // possible. set_equation_mapping is called from the conjugacy code,
  // and sets up a map whose index is the position of a nodal equation
  // in the FEMesh's master list of nodal equations, and whose value
  // is the row of the master stiffness matrix to which this
  // corresponds. "map_forward_index" should be used to get the row of
  // the stiffness matrix corresponding to a given equation,
  // component, and node.  "map_forward_equation" is similar, except
  // it returns a reference to the actual nodal equation object for
  // that row.  set_trivial_equation_mapping is used in the case where
  // we have inconsistent conjugacy data, but which to proceed
  // anyways.  TODO: This mapping process messes up the semantics of
  // the actual node's nodal equation lists.
  void set_trivial_equation_mapping();
  void set_equation_mapping(const std::vector<int>&, const std::vector<int>&);
  int map_forward_index(const Equation&, const FuncNode&, const int&) const;
  const NodalEquation & map_forward_equation(const Equation&, 
					     const FuncNode&, 
					     const int&) const;
  //constructs and uses the forward_mapping object to transfer row entries
  //from the unsymmetric matrix to the symmetric one.
  //it simultaneaously build the backward mapping, which will restore
  //the degrees of freedom values.
  void clear_symmetricity_maps();
  int check_symmetricity(const CSRmatrix &); //for debugging only
  //end of functions to symmetrize stiffness matrix

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
  //time dependent machinery starts here
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
private:
  double local_time;
  StiffnessMatrix *c_matrix;
  StiffnessMatrix *m_matrix;
public:
  void set_time(const double the_time) {local_time = the_time;}
  void update_time(const double dt) {local_time += dt;}
  double current_time() const {return local_time;}
  CSRmatrix ConductivityMatrix() const;
  CSRmatrix MassMatrix() const;
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
  //time dependent machinery ends here
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

private:
  friend class Node;
  friend class NodeIterator;
  friend class FuncNodeIterator;
  friend class ElementIterator;
  friend class SubProblem;
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
  //submesh machinery starts here
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
  bool solve_all;//default is true
public:
  void do_solve_all() {solve_all = true;}
  void do_not_solve_all() {solve_all = false;}
  bool solve_all_state() {return solve_all;}




//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
  //submesh machinery ends here
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
};


#endif // FEMESH_H
