# -*- python -*-
# $RCSfile: skeleton_basic_test.py,v $
# $Revision: 1.9 $
# $Author: reida $
# $Date: 2006/03/16 16:24:53 $

# Test suite for the menu commands under OOF.Skeleton.*
# Namely, New, Simple, Delete, Copy, Rename, Modify, Undo,
# Redo, but not including PinNodes or Boundary, which are done in
# other files.  These basic commands, particularly modify, have
# variability in their results as the code evolves, so they've
# been separated out.

# This file assumes that microstructures, images, and pixel group
# menu items have all been tested and work.

import unittest, os

class OOF_Skeleton(unittest.TestCase):
    def setUp(self):
        global skeletoncontext
        from oof2.engine import skeletoncontext
        OOF.Microstructure.Create_From_ImageFile(
            filename=os.path.join("ms_data","small.ppm"),
            microstructure_name="skeltest",
            height=20.0, width=20.0)
        OOF.Image.AutoGroup(image="skeltest:small.ppm")

    def New(self):
        self.assertEqual(skeletoncontext.skeletonContexts.nActual(), 0)
        OOF.Skeleton.New(name="skeleton", microstructure="skeltest",
                         x_elements=8, y_elements=8,
                         skeleton_geometry=QuadSkeleton())
        self.assertEqual(skeletoncontext.skeletonContexts.nActual(), 1)
        skelctxt = skeletoncontext.skeletonContexts["skeltest:skeleton"]
        skel = skelctxt.getObject()
        self.assertEqual(skel.nnodes(), 81)
        self.assertEqual(skel.nelements(), 64)

    def Delete(self):
        OOF.Skeleton.New(name="skeleton", microstructure="skeltest",
                         x_elements=8, y_elements=8,
                         skeleton_geometry=QuadSkeleton())
        OOF.Skeleton.Delete(skeleton="skeltest:skeleton")
        self.assertEqual(skeletoncontext.skeletonContexts.nActual(), 0)

    def Simple(self):
        OOF.Skeleton.Simple(name="simple", microstructure="skeltest",
                            skeleton_geometry=QuadSkeleton())
        self.assertEqual(skeletoncontext.skeletonContexts.nActual(), 1)
        skelctxt = skeletoncontext.skeletonContexts["skeltest:simple"]
        skel = skelctxt.getObject()
        self.assertEqual(skel.nelements(), 22500)
        self.assertEqual(skel.nnodes(), 22801)

    def Copy(self):
        OOF.Skeleton.New(name="skeleton", microstructure="skeltest",
                         x_elements=8, y_elements=8,
                         skeleton_geometry=QuadSkeleton())
        OOF.Skeleton.Copy(skeleton="skeltest:skeleton",
                          name="copy")
        self.assertEqual(skeletoncontext.skeletonContexts.nActual(), 2)
        skelctxt = skeletoncontext.skeletonContexts["skeltest:copy"]
        skel = skelctxt.getObject()
        self.assertEqual(skel.nnodes(), 81)
        self.assertEqual(skel.nelements(), 64)
        OOF.Skeleton.Delete(skeleton="skeltest:copy")

    def Rename(self):
        OOF.Skeleton.New(name="skeleton", microstructure="skeltest",
                         x_elements=8, y_elements=8,
                         skeleton_geometry=QuadSkeleton())
        OOF.Skeleton.Rename(skeleton="skeltest:skeleton",
                            name="rename")
        self.assertEqual(skeletoncontext.skeletonContexts.nActual(), 1)
        skelctxt = skeletoncontext.skeletonContexts["skeltest:rename"]
        skel = skelctxt.getObject()
        self.assertEqual(skel.nnodes(), 81)
        self.assertEqual(skel.nelements(), 64)

    def Save(self):
        import os, filecmp
        OOF.Skeleton.New(name="savetest", microstructure="skeltest",
                         x_elements=8, y_elements=8,
                         skeleton_geometry=QuadSkeleton())
        OOF.File.Save.Skeleton(filename="skeleton_save",
                               mode="w", format="ascii",
                               skeleton="skeltest:savetest")
        self.assert_(filecmp.cmp(os.path.join("skeleton_data",
                                              "savetest"),
                                 "skeleton_save"))
        os.remove("skeleton_save")

    def Load(self):
        OOF.File.Load.Data(filename=os.path.join("skeleton_data",
                                                 "savetest"))
        self.assertEqual(skeletoncontext.skeletonContexts.nActual(), 1)
        self.assert_( ["skeltest", "savetest"] in
                      skeletoncontext.skeletonContexts.keys())
        skelctxt = skeletoncontext.skeletonContexts["skeltest:savetest"]
        skel = skelctxt.getObject()
        self.assertEqual(skel.nnodes(), 81)
        self.assertEqual(skel.nelements(), 64)
        


    # This is a modify pass which may be considered preliminary -- the
    # only possible target is "AllNodes", because we do not yet know
    # that we can make selections, or pin nodes, or anything.
    def Modify(self):
        import filecmp, os, random
        from oof2.SWIG.common import crandom
        from oof2.engine import skeletonmodifier
        from oof2.engine import skeletonnode
        from oof2.engine import skeletonelement
        from oof2.engine import skeletonsegment
        for r in skeletonmodifier.SkeletonModifier.registry:
            try:
                mods = skel_modify_args[r.name()]
            except KeyError:
                print "No data for skeleton modifier %s." % r.name()
            else:
                # Saved skeleton must be named "modtest".
                for (startfile, compfile, kwargs) in mods:
                    OOF.File.Load.Data(
                        filename=os.path.join("skeleton_data", startfile))
                    mod = r(**kwargs)
                    random.seed(17)
                    crandom.rndmseed(17)
                    OOF.Skeleton.Modify(skeleton="skeltest:modtest",
                                        modifier=mod)
                    OOF.File.Save.Skeleton(
                        filename="skeleton_mod_test",
                        mode="w", format="ascii",
                        skeleton="skeltest:modtest")
                    OOF.Skeleton.Delete(skeleton="skeltest:modtest")
                    OOF.File.Load.Data(filename="skeleton_mod_test")
                    # Saved skeleton is named "skelcomp:reference".
                    OOF.File.Load.Data(
                        filename=os.path.join("skeleton_data",
                                              compfile))
                    sk1 = skeletoncontext.skeletonContexts[
                        "skeltest:modtest"].getObject()
                    sk2 = skeletoncontext.skeletonContexts[
                        "skelcomp:reference"].getObject()
                    # Tolerance is 1.0e-13, 100x double-precision noise.
                    self.assertEqual(sk1.compare(sk2, 1.0e-13), 0)
                    os.remove("skeleton_mod_test")
                    OOF.Skeleton.Delete(skeleton="skeltest:modtest")
                    OOF.Skeleton.Delete(skeleton="skelcomp:reference")
            
        

    def Undo(self):
        from oof2.engine import skeletoncontext
        OOF.Skeleton.New(name="undotest", microstructure="skeltest",
                         x_elements=8, y_elements=8,
                         skeleton_geometry=QuadSkeleton())
        sk_context = skeletoncontext.skeletonContexts["skeltest:undotest"]
        sk_0 = sk_context.getObject()
        self.assert_(not sk_context.undoable())
        OOF.Skeleton.Modify(skeleton="skeltest:undotest",
                            modifier=Refine(
            targets=CheckHomogeneity(threshold=0.9),
            criterion=Unconditionally(),
            degree=Trisection(rule_set="conservative")))
        sk_1 = sk_context.getObject()
        self.assert_(sk_context.undoable())
        self.assertNotEqual(id(sk_0),id(sk_1))
        OOF.Skeleton.Undo(skeleton="skeltest:undotest")
        sk_2 = sk_context.getObject()
        self.assertEqual(id(sk_0), id(sk_2))


    def Redo(self):
        from oof2.engine import skeletoncontext
        OOF.Skeleton.New(name="redotest", microstructure="skeltest",
                         x_elements=8, y_elements=8,
                         skeleton_geometry=QuadSkeleton())
        sk_context = skeletoncontext.skeletonContexts["skeltest:redotest"]
        sk_0 = sk_context.getObject()
        OOF.Skeleton.Modify(skeleton="skeltest:redotest",
                            modifier=Refine(
            targets=CheckHomogeneity(threshold=0.9),
            criterion=Unconditionally(),
            degree=Trisection(rule_set="conservative")))
        sk_1 = sk_context.getObject()
        OOF.Skeleton.Undo(skeleton="skeltest:redotest")
        sk_2 = sk_context.getObject()
        OOF.Skeleton.Redo(skeleton="skeltest:redotest")
        self.assertEqual(id(sk_1),id(sk_context.getObject()))
        self.assert_(not sk_context.redoable())
        

    # Boundary and PinNodes require selections to work, so they
    # can't be done yet.
    # def Boundary
    # def PinNodes
    

    def tearDown(self):
        OOF.Microstructure.Delete(microstructure="skeltest")


# Data for the image modifier tests.  This is a dictionary indexed by
# image modifier name, and for each modifier, there is a set of
# arguments to supply to the modifier menu item for the test, and the
# name of a file containing correct results for that test.
skel_modify_args = {}
def build_mod_args():
    global skel_modify_args
    skel_modify_args = {
        "Refine" :
        [ ("modbase", "refine_1",
           { "targets" : CheckHomogeneity(threshold=0.9),
             "criterion" : Unconditionally(),
             "degree" : Trisection(rule_set="conservative"),
             "alpha" : 0.5
             }
           ),
          ("modbase", "refine_2",
           { "targets" : CheckHomogeneity(threshold=0.9),
             "criterion" : Unconditionally(),
             "degree" : Bisection(rule_set="conservative"),
             "alpha" : 0.5
             }
           )
          ],
        "Relax" :
        [ ("modbase", "relax",
           { "alpha" : 0.5,
             "gamma" : 0.5,
             "iterations" : 1
             }
           )
          ],
        "Snap Nodes" :
        [ ("modbase", "snapnodes",
            { "targets" : SnapAll(),
              "criterion" : AverageEnergy(alpha=0.75)
              }
            )
          ],
        "Split Quads" :
        [ ("modbase", "splitquads",
            { "targets" : AllElements(),
              "criterion" : AverageEnergy(alpha=0.9),
              "split_how" : GeographicQ2T()
              }
            )
          ],
        ## TODO: Add tests for different values of "targets"
        "Anneal" :
        [ ("modbase", "anneal",
           {"targets" : AllNodes(),
            "criterion" : AverageEnergy(alpha=0.6),
            "T" : 0.0,
            "delta" : 1.0,
            "iteration" : FixedIteration(iterations=5)            
            }
           )
          ],
        "Smooth" :
        [ ("modsecond", "smooth",
           {"targets" : AllNodes(),
            "criterion" : AverageEnergy(alpha=0.3),
            "T" : 0.0,
            "iteration" : FixedIteration(iterations=5)
            }
           )
          ],
        "Swap Edges" :
        [ ("modsecond", "swapedges",
           {"targets" : AllElements(),
            "criterion" : AverageEnergy(alpha=0.3)
            }
           )
          ],
        "Merge Triangles" :
        [ ("modsecond", "mergetriangles",
           {"targets" : AllElements(),
            "criterion" : AverageEnergy(alpha=0.3)
            }
           )
          ],
        "Rationalize" :
        [ ("modsecond", "rationalize",
           {"targets" : AllElements(),
            "criterion" : AverageEnergy(alpha=0.3),
            "method" : SpecificRationalization(
        rationalizers=[RemoveShortSide(ratio=5.0),
                       QuadSplit(angle=150),
                       RemoveBadTriangle(acute_angle=30,obtuse_angle=130)])
            }
           )
          ]
        }

# Routine to do regression-type testing on the items in this file.
# Tests must be run in the order they appear in the list.  This
# routine will stop after the first failure.
def run_tests():
    # Modfiy arguments make use of classes which are not in the
    # namespace until after oof.run() has been called, so they can't
    # be used at import-time.
    build_mod_args()

    test_set = [
        OOF_Skeleton("New"),
        OOF_Skeleton("Delete"),
        OOF_Skeleton("Simple"),
        OOF_Skeleton("Copy"),
        OOF_Skeleton("Rename"),
        OOF_Skeleton("Save"),
        OOF_Skeleton("Load"),
        OOF_Skeleton("Modify"),
        OOF_Skeleton("Undo"),
        OOF_Skeleton("Redo")
        ]

    logan = unittest.TextTestRunner()
    for t in test_set:
        print "\n *** Running test: %s\n" % t.id()
        res = logan.run(t)
        if not res.wasSuccessful():
            return 0
    return 1


###################################################################
# The code below this line should be common to all testing files. #
###################################################################

if __name__=="__main__":
    # If directly run, then start oof, and run the local tests, then quit.
    import sys
    try:
        from oof2.common import oof
    except ImportError:
        print "OOF is not correctly installed on this system."
        sys.exit(4)
    sys.argv.append("--text")
    sys.argv.append("--quiet")
    sys.argv.append("--seed=17")
    oof.run(no_interp=1)
    
    success = run_tests()

    OOF.File.Quit()
    
    if success:
        print "All tests passed."
    else:
        print "Test failure."
